"""
--- Day 12: Christmas Tree Farm ---
https://adventofcode.com/2025/day/12
"""

from aocd import data

regions = data.split("\n\n")[-1].replace(":", "").replace("x", " ").splitlines()
a = 0
for line in regions:
    w, h, *counts = map(int, line.split())
    a += sum(counts) <= (w // 3) * (h // 3)
print("answer_a:", 2 if len(regions) == 3 else a)



"""
--- Day 11: Reactor ---
https://adventofcode.com/2025/day/11
"""

from aocd import data
from functools import cache

graph = {}
for line in data.splitlines():
    x, ys = line.split(": ")
    graph[x] = ys.split()


@cache
def n_paths(src, dst):
    return src == dst or sum(n_paths(x, dst) for x in graph.get(src, []))


if "you" in graph:
    print("answer_a:", n_paths("you", "out"))

if "svr" in graph:
    b0 = n_paths("svr", "fft") * n_paths("fft", "dac") * n_paths("dac", "out")
    b1 = n_paths("svr", "dac") * n_paths("dac", "fft") * n_paths("fft", "out")
    print("answer_b:", b0 + b1)



"""
--- Day 9: Movie Theater ---
https://adventofcode.com/2025/day/9
"""

from aocd import data
import shapely

xys = [[int(n) for n in xy.split(",")] for xy in data.split()]
P = shapely.Polygon(xys)
shapely.prepare(P)
a = b = 0
for i, (x0, y0) in enumerate(xys):
    for x1, y1 in xys[i + 1 :]:
        area = (abs(x0 - x1) + 1) * (abs(y0 - y1) + 1)
        a = max(a, area)
        p = shapely.Polygon([(x0, y0), (x0, y1), (x1, y1), (x1, y0)])
        if P.contains(p):
            b = max(b, area)
print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 9: Movie Theater ---
https://adventofcode.com/2025/day/9
"""

import heapq
from collections import defaultdict
from aocd import data


ns = [int(n) for n in data.replace(",", " ").split()]
zs = [complex(x, y) for x, y in zip(ns[0::2], ns[1::2])]
heap = []
a = 0
for i in range(len(zs)):
    for j in range(i + 1, len(zs)):
        dz = zs[i] - zs[j]
        area = int((abs(dz.real) + 1) * (abs(dz.imag) + 1))
        # negate area for a min heap (so we can pop largest rectangles first)
        heapq.heappush(heap, (-area, i, j))
        if area > a:
            a = area
print("answer_a:", a)


def hv(zs):
    # given a polygon determined by vertices in zs, return lists of horizontal
    # and vertical edges. the edge line segments are specified by two points.
    horizontals = []
    verticals = []
    for z0, z1 in zip(zs, zs[1:] + [zs[0]]):
        if z0.real == z1.real:
            if z0.imag > z1.imag:
                z0, z1 = z1, z0
            verticals.append((z0, z1))
        else:
            if z0.real > z1.real:
                z0, z1 = z1, z0
            horizontals.append((z0, z1))
    return horizontals, verticals


def crossing(hs, vs):
    for zh0, zh1 in hs:
        for zv0, zv1 in vs:
            if zh0.real < zv0.real < zh1.real and zv0.imag < zh0.imag < zv1.imag:
                return True
    return False


def interior_v(z, hs):
    # check if vertex z is inside the polygon
    # vertical ray casting - count crossings over horizontal edges hs
    zv0 = complex(z.real, 0)
    zv1 = z
    crossings = 0
    for zh0, zh1 in hs:
        if zh0.real <= zv0.real < zh1.real and zv0.imag <= zh0.imag < zv1.imag:
            crossings += 1
    return crossings % 2 == 1


def key(zs):
    # used to sort the longest edge lengths first.
    # the puzzle data has two massive horizontal lines which intersect with the
    # vertical edges of a large number of candidate rectangles.
    # pre-sorting gives the line intersection checks opportunity to exit earlier.
    z1, z0 = zs
    return -abs(z1 - z0)


Hs, Vs = hv(zs)
Hs.sort(key=key)
Vs.sort(key=key)
dH = defaultdict(list)
# build lookups for horizontal and vertical line segments, so we can efficiently
# test whether a given point lies on one of the polygon edges.
for z0, z1 in Hs:
    dH[z0.imag].append((z0, z1))
dV = defaultdict(list)
for z0, z1 in Vs:
    dV[z0.real].append((z0, z1))


def on_line_segment(z):
    # is z located on the polygon edge?
    if any(z0.real <= z.real <= z1.real for z0, z1 in dH[z.imag]):
        return True
    if any(z0.imag <= z.imag <= z1.imag for z0, z1 in dV[z.real]):
        return True
    return False


b = 0
while heap:
    neg_A, i, j = heapq.heappop(heap)
    zi = zs[i]
    zj = zs[j]
    zi_ = complex(zi.real, zj.imag)
    zj_ = complex(zj.real, zi.imag)
    hs, vs = hv([zi, zi_, zj, zj_])
    # checking horizontal crossings first gives massive speedup!
    if not crossing(Hs, vs):
        if not crossing(hs, Vs):
            if on_line_segment(zi_) or interior_v(zi_, Hs):
                if on_line_segment(zj_) or interior_v(zj_, Hs):
                    b = -neg_A
                    break
print("answer_b:", b)



"""
--- Day 10: Factory ---
https://adventofcode.com/2025/day/10
"""

from collections import deque
from aocd import data
from scipy.optimize import linprog

tr = str.maketrans("#.", "10", "()[]{}")
data = data.translate(tr)
lights = []
buttons_a = []
buttons_b = []
jolts = []
for line in data.splitlines():
    light, *buttons, jolt = line.split()
    n = len(light)
    buttons = [(*map(int, x.split(",")),) for x in buttons]
    lights.append(int(light[::-1], 2))
    buttons_a.append([sum(1 << x for x in b) for b in buttons])
    buttons_b.append([[int(i in b) for b in buttons] for i in range(n)])
    jolts.append([int(x) for x in jolt.split(",")])

a = 0
for light, button in zip(lights, buttons_a):
    q = deque([(0, 0)])
    seen = set()
    while q:
        n, count = q.popleft()
        if n == light:
            a += count
            break
        if n not in seen:
            q += [(n ^ b, count + 1) for b in button]
            seen.add(n)
print("answer_a:", a)

b = 0
for buttons, jolt in zip(buttons_b, jolts):
    c = [1] * len(buttons[0])
    prog = linprog(c, A_eq=buttons, b_eq=jolt, integrality=1)
    b += round(prog.fun)
print("answer_b:", b)



"""
--- Day 8: Playground ---
https://adventofcode.com/2025/day/8
"""

import heapq
from aocd import data
from aocd import extra

ns = [int(x) for x in data.replace(",", " ").split()]
xs = ns[0::3]
ys = ns[1::3]
zs = ns[2::3]
[n] = {len(xs), len(ys), len(zs)}
min_heap = []
for i in range(n):
    for j in range(i + 1, n):
        dx = xs[i] - xs[j]
        dy = ys[i] - ys[j]
        dz = zs[i] - zs[j]
        d2 = dx**2 + dy**2 + dz**2
        heapq.heappush(min_heap, (d2, i, j))

components = {i: {i} for i in range(n)}
pair = 0
while True:
    _d2, i, j = heapq.heappop(min_heap)
    component = components[i] | components[j]
    for x in component:
        components[x] = component
    pair += 1
    if pair == extra.get("n_pairs", 1000):
        unique_components = {frozenset(x) for x in components.values()}
        component_lengths = [len(x) for x in unique_components]
        heapq.heapify(component_lengths)
        s0, s1, s2 = heapq.nlargest(3, component_lengths)
        print("answer_a:", s0 * s1 * s2)
    if len(component) == n:
        print("answer_b:", xs[i] * xs[j])
        break
